/*
 ============================================================================

  This file is part of the bbLean source code
  Copyright  2001-2004 grischka and The Blackbox for Windows Development Team

  http://bb4win.sourceforge.net/bblean
  http://sourceforge.net/projects/bb4win

 ============================================================================

  bbLean and bb4win are free software, released under the GNU General
  Public License (GPL version 2 or later), with an extension that allows
  linking of proprietary modules under a controlled interface. This means
  that plugins etc. are allowed to be released under any license the author
  wishes. For details see:

  http://www.fsf.org/licenses/gpl.html
  http://www.fsf.org/licenses/gpl-faq.html#LinkingOverControlledInterface

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
  for more details.

 ============================================================================
*/


///////////////////////////////////////////////////////////////////////////
#define strlwr _strlwr
#define stricmp _stricmp

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>

extern "C"
{
struct _list   { struct _list *next; void *v; };

void appendlist (void *a,void *e);  // 
void conslist (void *a,void *e);  // 
void delete_item(void *a,void *e);  //
void freeall(void *p);  // 
void *new_node(void *p);  // 

#define dolist(_e,_l) for (_e=_l;_e;_e=_e->next)
};

int MBoxErrorValue(LPCSTR szValue);
LPCSTR ReadValue(LPCSTR fp, LPCSTR keyword, LPLONG ptr=NULL);

///////////////////////////////////////////////////////////////////////////



//inline void *m_alloc(unsigned s) { return malloc(s); }
//inline void *c_alloc(unsigned s) { return calloc(1,s); }
//inline void m_free(void *v)      { free(v); }

//===========================================================================
// tiny cache reader: first checks, if the file had already been read, if not,
// reads the file into a malloc'ed buffer, then for each line separates
// keyword from value, cuts off leading and trailing spaces, strlwr's the
// keyword and adds both to a list of below defined structures, where k is
// the offset to the start of the value-string. Comments or other non-keyword
// lines have a "" keyword and the line as value.

// added checking for external updates by the user.

// added *wild*card* processing, it first looks for an exact match, if it
// cant find any, returns the first wildcard value, that matches, or null,
// if none...

//===========================================================================
// structures

struct lin_list
{   struct lin_list *next;
	int k;
	int o;
	unsigned hash;
	char is_wild;
	char ext_upd;
	char str[1];
};

static struct fil_list
{   struct fil_list *next;
	struct lin_list *lines;
	struct _list *wild;
	int max_len;
	FILETIME ft;
	char time_flg;
	char fname[1];
} *rc_files = NULL;


//===========================================================================
// used to remove old stylefile after style-changes

void delete_from_list(struct fil_list *fl)
{
	freeall(&fl->lines);
	freeall(&fl->wild);
	delete_item(&rc_files, fl);
}

//===========================================================================
// helpers

unsigned hash(char *s) {
	unsigned h; unsigned char c;
	for (h=0;0!=(c=*s++);)
	{
	   if ((h^=c)&1) h^=0xedb88320;
	   h>>=1;
	}
	return h;
}

static struct lin_list *make_line (const char *key, int k, const char *val)
{
	struct lin_list *tl; char *p; int o;
	o = strlen(val);
	tl=(struct lin_list*)malloc(sizeof(*tl) + (k+1)*2 + o);
	p=tl->str;
	memcpy(p, key, k); p[k] = 0;
	tl->hash = hash(strlwr(p)); // strlwr'd keyword for compare
	tl->k = ++k;
	memcpy(p+k, val, ++o);  // this is the value
	tl->o = o = k+o;
	memcpy(p+o, key, --k);  // the keyword again, not strlwr'd for writing
	p[o+k] = 0;
	tl->next = NULL;
	tl->ext_upd = tl->is_wild = 0;
	return tl;
}

// simple wildcard pattern matcher
static int match(char *str, char *pat)
{
	int n, m = 0;
	for (;;)
	{   char s = *str, p = *pat;
		if ('*' == p)
		{   if (s && 0 != (n = match(str+1, pat)))
				return m+n;
			++pat;
			continue;
		}
		if (0 == s)
		{
			if (s != p) return 0;
			return m;
		}
		if ('?' != p)
		{
			if (s != p) return 0;
			++m;
		}
		++str, ++pat;
	}
}

static struct lin_list *search_line(struct lin_list *tl, char *s, int k, struct _list *wild)
{
	unsigned h = hash(s);
	while (tl)
	{
		if (tl->hash==h && tl->k==k && 0==memcmp(tl->str, s, k))
			return tl;
		tl=tl->next;
	}
	struct lin_list *wl = NULL; int n, m = 0;
	while (wild)
	{
		tl = (struct lin_list *)wild->v;
		if (m < (n = match(s, tl->str)))
			wl = tl, m = n;
		wild=wild->next;
	}
	return wl;
}

int getftime(const char *fn, FILETIME *ft)
{
	WIN32_FIND_DATA data_bb;
	HANDLE h = FindFirstFile(fn, &data_bb);
	if (INVALID_HANDLE_VALUE==h) return 0;
	FindClose(h);
	*ft = data_bb.ftLastWriteTime;
	return 1;
}

int cmpftime(const char *fn, FILETIME *ft0)
{
	FILETIME ft;
	return getftime(fn, &ft) == 0 || CompareFileTime(&ft, ft0) > 0;
}

char *read_file_into_buffer(const char *fname)
{
	FILE *fp; char *buf; int k;
	if (NULL == (fp=fopen(fname,"rb")))
		return NULL;

	fseek(fp,0,SEEK_END);
	k = ftell (fp);
	fseek (fp,0,SEEK_SET);

	buf=(char*)malloc(k+1);
	fread (buf, 1, k, fp);
	fclose(fp);

	buf[k]=0;
	return buf;
}

//===========================================================================
// searches for the filename and, if not found, builds a new line-list

static struct fil_list *read_file(const char *fname)
{
	struct lin_list **slp, *sl;
	struct fil_list **flp, *fl, *fl0;
	char *buf, *p, *d, *s, c; int k;

	// first check, if the file has already been read
	for (flp = &rc_files; NULL!=(fl=*flp); flp = &fl->next)
	{
		if (0==stricmp(fname, fl->fname))
		{   //and if it was not updated from outside...
//#if 1
#if 0
			if (fl->time_flg)
			{
				if (BBhwnd)
				{
					SetTimer(BBhwnd, 2, 200, NULL);
					fl->time_flg = 0;
				}
				if (cmpftime(fname, &fl->ft))
					break;
			}
#else
			if (cmpftime(fname, &fl->ft))
				break;
#endif
			return fl; //... return cached line list.
		}
	}
	fl0 = fl; //not null, it externally updated

	// allocate a new file structure
	fl = (struct fil_list*)calloc(1, sizeof(*fl) + strlen(fname));
	conslist(&rc_files, fl);
	strcpy(fl->fname, fname);
	fl->time_flg = 1;

	if (NULL==(buf=read_file_into_buffer(fname)))
		goto end_1;

	//set timestamp
	getftime(fname, &fl->ft);

	slp = &fl->lines; p = buf;
	for (;;)
	{
		//skip leading spaces
		for (; 0!=(c=*p) && 10!=c && (unsigned char)c<=32; p++);
		//EOB reached ?
		if (0==c) break;

		//find end of line
		for (s=p; 0!=(c=*p) && 10!=c; p++);

		//cut off trailing spaces
		for (d=p; d>s && (unsigned char)d[-1] <= 32; d--);
		*d=0;

		//ready for next line
		if (0!=c) p++;

		//check for comments...
		if (*s == '#' || *s == '!' || NULL==(d=(char*)memchr(s,':',d-s)))
		{
			k=0, d=s;
		}
		else
		//skip spaces between key and value
		{
			//replace tabs
			for (char *t = s; 0!=(c=*t); t++) if ((unsigned char)c<32) *t=32;
			k=d-s+1;
			while (*++d == 32);
		}

		//put it into the line structure
		sl = make_line(s, k, d);
		if (k)
		{
			if (k>fl->max_len) fl->max_len = k;
			//check, it the key contains a wildcard
			if (memchr(s, '*', k))
			{
				// add it to the wildcard - list
				appendlist(&fl->wild, new_node(sl));
				sl->is_wild = true;
			}
#if 1
			if (fl0)
			{   //if this is a externally updated file, look, if the value
				//was present
				struct lin_list *tl = search_line(fl0->lines, sl->str, sl->k, NULL);
				// and if it had been changed, set the extern_updated flag
				if (tl && 0!=strcmp(tl->str+tl->k, sl->str+sl->k))
					sl->ext_upd = 1;
			}
#endif
		}
		//append it to the list
		slp = &(*slp=sl)->next;
	}
	free(buf);
end_1:
	if (fl0)
	{
		delete_from_list(fl0);
	}
	return fl;
}

//===========================================================================
// Function: ReadValue
// Purpose: Searches the given file for the supplied keyword and returns a
// pointer to the value - string
// In: LPCSTR fname = String containing the name of the file to be opened
// In: LPCSTR keyword = String containing the keyword to be looked for
// Out: LPSTR = Pointer to the value string of the keyword
//===========================================================================
static int found_value;

LPCSTR ReadValue(LPCSTR fname, LPCSTR keyword, LONG *ptr)
{
	struct lin_list *tl; struct _list *wild; char s[128]; int k;
	memcpy(s, keyword, k = 1 + strlen(keyword)); strlwr(s);

	if (ptr)
	{
		if (0 == *ptr)
			tl = read_file(fname)->lines;
		else
			tl = **(struct lin_list ***)ptr;
		wild = NULL;
	}
	else
	{
		struct fil_list *fl = read_file(fname);
		tl = fl->lines;  wild = fl->wild;
	}
	if (0 == (found_value=(NULL!=(tl=search_line(tl, s, k, wild)))))
		return NULL;

	if (tl->is_wild)
		found_value = 2;

	if (ptr)
		*(struct lin_list ***)ptr = &tl->next;

	tl->ext_upd=0;
	return tl->str+tl->k;
}

//===========================================================================
FILE *create_rcfile(const char *fname)
{
	FILE *fp;
	if (NULL==(fp=fopen(fname, "wt")))
	{
		char buffer[MAX_PATH];
		sprintf(buffer,"Could not write to '%s'.", fname);
		MBoxErrorValue(buffer);
	}
	return fp;
}

void write_rcfile(const char *fname, struct fil_list *fl)
{
	FILE *fp;
	if (NULL==(fp=create_rcfile(fname)))
		return;

	struct lin_list *tl;
	dolist (tl, fl->lines)
	{
		char *s = tl->str + tl->k;
		int ml = fl->max_len | 3;
		if (0 == *tl->str) //comment
			fprintf (fp, "%s\n", s);
		else
			fprintf (fp, "%-*s %s\n", ml, tl->str+tl->o, s);
	}
	fclose(fp);
}

//===========================================================================
// Put keyword & value into the above defined struct, search for the
// key in the list, replace, if found and the value did change, or
// append, if not found. Write to file on changes.

void WriteValue(LPCSTR fname, LPCSTR keyword, LPCSTR value, bool saveToFile)
{
	struct lin_list *tl, **slp, *sl; char buff[80]; int k;
	struct fil_list *fl = read_file(fname);
	memcpy(buff, keyword, 1 + (k = strlen(keyword)));
	strlwr(buff);

	for (slp = &fl->lines; NULL!=(sl=*slp); slp = &(*slp)->next)
	{
		if (sl->k==1+k && 0==memcmp(sl->str, buff, k) && k)
		{
			//if it didn't change or has the extern_updated flag set, quit
			if (value && (0==strcmp(sl->str + sl->k, value) || sl->ext_upd))
				return;

			*slp = sl->next; free(sl);
			break;
		}
	}

	if (value)
	{
		tl = make_line(keyword, k, value);
		tl->next = *slp;
		*slp = tl;
		if (k > fl->max_len) fl->max_len = k;
	}

	if (saveToFile)
	{
		write_rcfile(fl->fname, fl);
		getftime(fl->fname, &fl->ft);
	}
}

//===========================================================================
// API functions

void WriteString(LPCSTR fileName, LPCSTR keyword, LPCSTR value)
{
	WriteValue(fileName, keyword, value, true);
}

/*
void WriteBool(LPCSTR fileName, LPCSTR keyword, bool value)
{
	WriteString(fileName, keyword, value ? "true" : "false");
}
*/

void WriteInt(LPCSTR fileName, LPCSTR keyword, int value)
{
	char buff[32];
	sprintf(buff, "%d", value);
	WriteString(fileName, keyword, buff);
}

//===========================================================================
// Function: ReadBool
// Purpose: Read's a string in the given file and outputs the result as a boolean
// In: LPCSTR fp = File to be searched
// In: LPCSTR szString = Keyword to search for
// In: bool szDefault = Default value to return if not found
// Out: bool = Returns the value of the found bool
//===========================================================================

/*
bool ReadBool(LPCSTR fp, LPCSTR szString, bool szDefault)
{
	LPCSTR vstr = ReadValue(fp, szString);
	if (vstr)
	{
		if (!stricmp(vstr, "true"))  return true;
		if (!stricmp(vstr, "false")) return false;
	}
	return szDefault;
}
*/

//===========================================================================
// Function: ReadInt
// Purpose: Searches given file for int value
// In: LPCSTR fp = File to search
// In: LPCSTR szString = Keyword to look for
// In: int szDefault = Default value to to return if not found
// Out: int = Value found
//===========================================================================

int ReadInt(LPCSTR fp, LPCSTR szString, int szDefault)
{
	LPCSTR vstr = ReadValue(fp, szString);
	if (vstr) return atoi(vstr);
	return szDefault;
}

//===========================================================================
// Function: ReadString
// Purpose: Searches given file for string
// In: LPCSTR fp = File to search
// In: LPCSTR szString = String to return value of
// In: LPSTR szDefault = Pointer to value to return if not found
// Out: LPSTR = Pointer to found string
//===========================================================================

/*
LPCSTR ReadString(LPCSTR fp, LPCSTR szString, LPCSTR szDefault)
{
	LPCSTR vstr = ReadValue(fp, szString);
	if (vstr) return vstr;
	return szDefault;
}
*/

//===========================================================================
// Function: MBoxErrorValue
// Purpose: Gives a message box proclaming a given value
// In: LPCSTR = value
// Out: int = return value of messagebox
//===========================================================================

int bb_MessageBox(int flg, LPCSTR msg)
{
	const char *p, *caption = "Blackbox"; char buffer[128];
	if ('#' == msg[0] && NULL != (p = strchr(msg+1, msg[0])))
	{
		caption = (const char*)memcpy(buffer, msg+1, p - msg - 1);
		buffer[p - msg - 1] = 0; msg = p+1;
	}
	//return MessageBox (BBhwnd, msg, caption,
	return MessageBox (NULL, msg, caption,
		flg | MB_TOPMOST | MB_SETFOREGROUND | MB_APPLMODAL);
}

int MBoxErrorValue(LPCSTR szValue)
{
	char temp[1024];
	sprintf(temp, "Error: %s\t", szValue);
	return bb_MessageBox(MB_OK, temp);
}

