/*
 ============================================================================

  This file is part of the bbLean source code
  Copyright  2001-2004 grischka and The Blackbox for Windows Development Team

  http://bb4win.sourceforge.net/bblean
  http://sourceforge.net/projects/bb4win

 ============================================================================

  bbLean and bb4win are free software, released under the GNU General
  Public License (GPL version 2 or later), with an extension that allows
  linking of proprietary modules under a controlled interface. This means
  that plugins etc. are allowed to be released under any license the author
  wishes. For details see:

  http://www.fsf.org/licenses/gpl.html
  http://www.fsf.org/licenses/gpl-faq.html#LinkingOverControlledInterface

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
  for more details.

 ============================================================================
*/

// TINYLIST.C - some simple (and fast) list funcions

#ifdef __cplusplus
extern "C" {
#define NULL 0
#else
#define NULL (void*)0
#endif

//////////////////////////
/*
void *m_alloc(unsigned n);
void *c_alloc (unsigned x);
void m_free(void *p);
*/
#include <malloc.h>
//////////////////////////

unsigned int strlen(const char*);
char* strcpy(char*, const char*);

int imax(int a, int b) {
	return a>b?a:b;
}

int imin(int a, int b) {
	return a<b?a:b;
}

int iminmax(int a, int b, int c) {
	if (a<b) a=b;
	if (a>c) a=c;
	return a;
}

struct strlist { struct strlist *next; char str[1]; };

typedef struct _list { struct _list *next; void *v; } _list;

void remove_node (void *a0, void *e0)
{
	_list *a = (_list*)a0, *e = (_list*)e0;
	for (;a;a=a->next)
		if (a->next==e)
		{
			a->next=a->next->next;
			break;
		}
}

void delete_item(void *a,void *e)
{
	if (e)
	{
		remove_node(a, e);
		free(e);
	}
}

void freeall(void *p)
{   _list *s, *q;
	q=*(_list**)p;
	for (; q; q=(s=q)->next, free(s));
	*(_list**)p=q;
}

void revlist (void *d)
{   _list *a, *b, *c;
	a=*(_list**)d;
	for (b=NULL; a; c=a->next, a->next=b, b=a, a=c);
	*(_list**)d=b;
}

void appendlist (void *a0 ,void *e0)
{
	_list *a = (_list*)a0, *e = (_list*)e0;
	for (;a->next;a=a->next);
	a->next=e; e->next=NULL;
}

void conslist (void *a0, void *e0)
{
	_list *a = (_list*)a0, *e = (_list*)e0;
	e->next=a->next; a->next=e;
}

void *member(void *a0, void *e0)
{
	_list *a = (_list*)a0, *e = (_list*)e0;
	for (;a && a->next!=e; a=a->next);
	return a;
}

void *assoc(void *a0, void *e0)
{
	_list *a = (_list*)a0, *e = (_list*)e0;
	for (;a->next;a=a->next)
		if (e==a->next->v) return a;
	return NULL;
}

void delete_assoc(void *a,void *e)
{
	void *q, *p = assoc(a, e);
	if (NULL==p) return;
	q = *(void**)p;
	*(void**)p = *(void**)q;
	free(q);
}


void inslist (void *a0, void *e0, void *i0)
{
	_list *a = (_list*)a0, *e = (_list*)e0, *i = (_list*)i0, *x;
	if (e==i) return;
	remove_node(a, i);
	if (NULL==e) appendlist(a, i);
	else
	x=e->next, e->next=i, i->next=x;
}

struct strlist *newstrl(char *s)
{
	struct strlist *b;
	b = (struct strlist *)calloc(1, sizeof(*b)+strlen(s)+1);
	strcpy(b->str,s);
	return b;
};

void appendstr(struct strlist **p, char *s)
{
	appendlist(p, newstrl(s));
}


void *nth (void *v0, int n)
{
	_list *v = (_list*)v0;
	while (n && v)  n--, v=v->next;
	return v;
}

void *new_node(void *p)
{
	_list *n = (_list*)calloc(1, 2*sizeof(void*));
	n->v = p;
	return (void*)n;
}

int listlen(void *v0)
{
	_list *v = (_list*)v0; int i = 0;
	while(v) i++, v=v->next;
	return i;
}

char *new_str(const char *s) {
	return s ? strcpy((char*)malloc(strlen(s)+1), s) : NULL;
}

void free_str(char **s){
	if (*s) free(*s), *s=NULL;
}

/*----------------------------------------------------------------------------*/
#ifdef __cplusplus
};
#endif

