/*---------------------------------------------------------------------------------
 bbEdgeMenu ( 2004 Slade Taylor [bladestaylor@yahoo.com])
 ----------------------------------------------------------------------------------
 bbEdgeMenu is a plugin for Blackbox for Windows.  For more information,
 please visit [http://bb4win.org] or [http://sourceforge.net/projects/bb4win].
 ----------------------------------------------------------------------------------
 bbEdgeMenu is free software, released under the GNU General Public License,
 version 2, as published by the Free Software Foundation.  It is distributed
 WITHOUT ANY WARRANTY--without even the implied warranty of MERCHANTABILITY
 or FITNESS FOR A PARTICULAR PURPOSE.  Please see the GNU General Public
 License for more details:  [http://www.fsf.org/licenses/gpl.html].
 --------------------------------------------------------------------------------*/

#include "bbEdgeMenu.h"

//---------------------------------------------------------------------------------

bool WINAPI DllMain(HINSTANCE hinst, DWORD fdwReason, LPVOID lpvReserved)
{
	if (fdwReason == DLL_PROCESS_ATTACH)
	{
		DisableThreadLibraryCalls(hInstance = hinst);
		if (!(hCoreWnd = FindWindow("BlackBoxClass", "BlackBox"))) // get bb window
			hCoreWnd = FindWindow("xoblite", NULL);
		RCSettings();
		GetScreenSettings();
	}
	return true;
}

//---------------------------------------------------------------------------------

int beginPlugin(HINSTANCE hPluginInstance)
{
	WNDCLASS wc;
	ZeroMemory(&wc, sizeof(wc));
	wc.lpfnWndProc = PluginProc;
	wc.hInstance = hPluginInstance;
	wc.lpszClassName = pluginInfo(APPNAME);
	wc.hCursor = NULL;
	RegisterClass(&wc);

	// create window to catch display-change
	hPluginWnd = CreateWindowEx(
		WS_EX_TOOLWINDOW,
		pluginInfo(APPNAME),
		NULL,
		WS_POPUP,
		0, 0, 0, 0,
		NULL,
		NULL,
		hPluginInstance,
		NULL);

	BBThread = GetCurrentThreadId();  // get blackbox thread
	hMouseHook = SetWindowsHookEx(WH_MOUSE, MouseProc, hInstance, 0);
	return 0;
}

//---------------------------------------------------------------------------------

void endPlugin(HINSTANCE hPluginInstance)
{
	DestroyWindow(hPluginWnd);
	UnregisterClass(pluginInfo(APPNAME), hPluginInstance);
	if (hMouseHook)
		UnhookWindowsHookEx(hMouseHook);
}

//---------------------------------------------------------------------------------

LRESULT CALLBACK PluginProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	if (message == WM_DISPLAYCHANGE)
	{
		GetScreenSettings();  // get new screen settings on display-change
		return 0;
	}
	return DefWindowProc(hwnd, message, wParam, lParam);
}

//---------------------------------------------------------------------------------

void GetScreenSettings()
{
	int v_screen_x = 0;
	int v_screen_y = 0;
	int screen_width = GetSystemMetrics(SM_CXSCREEN);
	int screen_height = GetSystemMetrics(SM_CYSCREEN);
	int v_screen_width = GetSystemMetrics(SM_CXVIRTUALSCREEN);
	int v_screen_height = GetSystemMetrics(SM_CYVIRTUALSCREEN);	
	
	if ((v_screen_width > screen_width) || (v_screen_height > screen_height))
	{
		v_screen_x = GetSystemMetrics(SM_XVIRTUALSCREEN);
		v_screen_y = GetSystemMetrics(SM_YVIRTUALSCREEN);
	}
	else
	{
		v_screen_width  = screen_width;
		v_screen_height = screen_height;
	}
	
	ExclusionRect.top = v_screen_y + ExclusionRect.top;
	ExclusionRect.bottom = v_screen_y + v_screen_height - ExclusionRect.bottom;
	ExclusionRect.left = v_screen_x + ExclusionRect.left;
	ExclusionRect.right = v_screen_x + v_screen_width - ExclusionRect.right - 1;	
}

//---------------------------------------------------------------------------------

LRESULT CALLBACK MouseProc(int nCode, WPARAM wParam, LPARAM lParam)
{
	if (nCode == HC_ACTION) switch (wParam)
	{
		case WM_NCRBUTTONDOWN:
		case WM_RBUTTONDOWN:
			if (IsValid(lParam))
				return 1;
			break;

		case WM_NCRBUTTONUP:
		case WM_RBUTTONUP:
			if (IsValid(lParam))
			{
				PostMessage(hCoreWnd, BB_MENU, 0, 0);  // show desktop menu
				return 1;
			}
	}
	return CallNextHookEx(hMouseHook, nCode, wParam, lParam);
}

//---------------------------------------------------------------------------------

bool IsValid(LPARAM lParam)
{
	MOUSEHOOKSTRUCT *pData = (MOUSEHOOKSTRUCT*)lParam;
	if (GetWindowThreadProcessId(pData->hwnd, NULL) != BBThread)  // exclude bb windows
		if (!PtInRect(&ExclusionRect, pData->pt))  // if cursor in clickable area
			return true;
	return false;
}

//---------------------------------------------------------------------------------

void RCSettings()
{
	char rc_path[MAX_LINE_LENGTH];
	int i = GetModuleFileName(hInstance, rc_path, MAX_LINE_LENGTH);
	if (i > 3) { rc_path[i - 3] = 'r'; rc_path[i - 2] = 'c'; rc_path[i - 1] = 0; }

	RC_Int(rc_path,	"Left:",	&ExclusionRect.left,	1);
	RC_Int(rc_path,	"Right:",	&ExclusionRect.right,	1);
	RC_Int(rc_path,	"Top:",		&ExclusionRect.top,		0);
	RC_Int(rc_path,	"Bottom:",	&ExclusionRect.bottom,	0);
}

//---------------------------------------------------------------------------------

void RC_Int(char *path, char *label, long *target, int default_val)
{
	*target = ReadInt(path, label, default_val);
	WriteInt(path, label, *(int*)target);
}

//---------------------------------------------------------------------------------
